import RoundRobin
import Char
import Match
import CXXMatch as CXX
import M100Util

import Control.Monad
import Data.List
import Data.Maybe
import System.IO.SSC.Haskell98 as SSC
import Text.Printf
import System.Environment
import System.Console.GetOpt
import System.Exit

data Settings = Se
  { seHaskell :: Bool
  , seDefaultNamespace :: Namespace
  }

options :: [OptDescr (Settings -> Settings)]
options =
  [ Option "d" ["old-engine"] (NoArg (\se -> se{ seHaskell = True }))
      "use old simulator engine in Haskell"
  , Option "a" ["m2011a"] (NoArg (\se -> se{ seDefaultNamespace = M2011a }))
      "use swords from m2011-alpha by default"
  -- , Option "h" ["help"] (NoArg $ const HelpMode) "show this help"
  ]

handleOptions :: [String] -> IO ([Todo], Settings)
handleOptions args = case getOpt Permute options args of
  (ms, as, []) -> case as of
      [] -> return ([SelfMatch "player.txt"], se)
      [inp] -> return ([SelfMatch inp], se)
      _ -> return (cms, se)
        where
          cms = do
            (xn, xp) <- zip [0..] as
            (yn, yp) <- zip [0..] as
            guard $ xn < yn
            return $ CrossMatch (show xn ++ "-" ++ show yn) xp yp
    where
      se = foldr ($) Se{ seHaskell = False, seDefaultNamespace = M100 } ms
  (_, _, msgs) -> do
    mapM_ SSC.putStrLn msgs
    show_usage >> exitWith (ExitFailure 1)
  where
    show_usage = do
      SSC.putStrLn usage
      SSC.putStrLn "use -h for help"
    usage = "Usage: fake-marguered [OPTIONS...] [FILE]"
    show_help = SSC.putStr $ usageInfo usage options

data Todo = SelfMatch FilePath | CrossMatch String FilePath FilePath

main = do
  (todos, se) <- getArgs >>= handleOptions
  forM_ todos $ \todo -> case todo of
    SelfMatch fp -> doSelfMatch se fp
    CrossMatch sig x y -> doCrossMatch se sig x y

doSelfMatch :: Settings -> FilePath -> IO ()
doSelfMatch se fp = do
  (ignored, chs) <- getChars se fp
  out <- SSC.openFile "result.txt" SSC.WriteMode
  result <- rrWithOutput (seHaskell se) out chs
  let s_result = formatResult chs result ignored
  SSC.hPutStr out s_result
  SSC.hClose out
  SSC.writeFile "result.csv" $ formatResultAsCSV chs result
  SSC.writeFile "roundrobin.txt" s_result

doCrossMatch :: Settings -> String -> FilePath -> FilePath -> IO ()
doCrossMatch se sig x y = do
  (ignored0, chs0) <- getChars se x
  (ignored1, chs1) <- getChars se y
  out <- SSC.openFile ("cmresult" ++ sig ++ ".txt") SSC.WriteMode
  result <- cmWithOutput (seHaskell se) out chs0 chs1
  let s_result = formatResult2 chs0 chs1 result $ ignored0 ++ ignored1
  SSC.hPutStr out s_result
  SSC.hClose out
  SSC.writeFile ("roundrobin" ++ sig ++ ".txt") s_result

getChars :: Settings -> FilePath -> IO ([String], [Character])
getChars se fp = do
  (ignored, chs0) <- readCharsFromFileWithIgnored (seDefaultNamespace se) fp
  let chs = uniquifyNames chs0
  return (ignored, chs)

formatResult :: [Character] -> [[Result]] -> [String] -> String
formatResult chs rs ignored = "RESULT\n" ++ formatDescr chs ++ "\n" ++ formatTable rs ++ formatNotice ignored

formatResult2 :: [Character] -> [Character] -> [[Result]] -> [String] -> String
formatResult2 chs0 chs1 rs ignored =
  "RESULT\n" ++ formatDescr chs0 ++ "\n" ++ formatDescr chs1 ++ "\n" ++ formatTable rs ++ formatNotice ignored

formatDescr :: [Character] -> String
formatDescr chs = unlines $ zipWith descr [0..] chs
  where
    descr n ch = printf "%4d: %s" (n :: Int) (show ch)

formatTable :: [[Result]] -> String
formatTable rs = renderAsText $ makeStringTable (to_name rs) (to_name $ head rs) rs show ["", "", "", "", ""] count
  where
    to_name = zipWith (const . show) [0..]

formatNotice :: [String] -> String
formatNotice ignored
  | null ignored = ""
  | otherwise = unlines $ "\n":"お知らせ: 以下の各行が無視されました。":ignored

rrWithOutput :: Bool -> SSC.Handle -> [Character] -> IO [[Result]]
rrWithOutput use_old h chs = liftM (completeTable Draw negateResult) $ roundRobin (zip chs [0..]) $ \(y, yn) (x, xn) -> 
  fmap negateResult $ doMatch use_old h x xn y yn

cmWithOutput :: Bool -> SSC.Handle -> [Character] -> [Character] -> IO [[Result]]
cmWithOutput use_old h chs0 chs1 = crossMatch (zip chs0 [0..]) (zip chs1 [0..]) $ \(x, xn) (y, yn) -> 
  doMatch use_old h x xn y yn

doMatch :: Bool -> SSC.Handle -> Character -> Int -> Character -> Int -> IO Result
doMatch use_old h x xn y yn = do
  hPutStrLn h $ replicate 48 'z'
  hPutStrLn h $ "＠" ++ chName x ++ "　VS　" ++ chName y ++ " [" ++ show xn ++ "vs" ++ show yn ++ "]"
  hPutStrLn h ""
  gr <- do_match use_old h x y
  hPutStrLn h ""
  return $ grToResult gr
  where
    do_match True h x y = do
      hPutStr h $ unlines msgs
      grWinner gr `seq` return gr
      where
        (gr, msgs, _) = Match.match x y
    do_match False h x y = CXX.match h x y

data Result = Win | Lose | Draw
  deriving (Eq)

instance Show Result where
  show Win = "@@"
  show Lose = ". "
  show Draw = "++"

negateResult :: Result -> Result
negateResult r = case r of
  Win -> Lose
  Lose -> Win
  Draw -> Draw

totals :: [Result] -> [Int]
totals rs = map c [Win, Draw, Lose]
  where
    c r = length $ filter (==r) rs

count :: [Result] -> [String]
count rs = intercalate ["-"] $ map ((:[]) . show) $ totals rs

grToResult :: GameResult -> Result
grToResult gr = case grWinner gr of
  Nothing -> Draw
  Just Team0 -> Win
  Just Team1 -> Lose

formatResultAsCSV :: [Character] -> [[Result]] -> String
formatResultAsCSV cs tbl = renderCSV $ head0 : head1 : zipWith3 line [0..] cs tbl
  where
    head0 = "RESULT" : replicate 6 "" ++ names ++ [""]
    head1 = ["HP", "攻", "防", "速", "剣", "名前", "番号"] ++ map show nums ++ ["勝", "分", "負"]
    line n ch rs = map show [chHp ch, chAtk ch, chDef ch, chSpd ch] 
      ++ [concatMap swShortName $ chSwords ch, chName ch, show n] ++ map ((:[]) . head . show) rs ++ map show (totals rs)
    names = map chName cs
    nums = zipWith const [0..] cs

renderCSV :: [[String]] -> String
renderCSV tbl = unlines $ map tr_line tbl
  where
    tr_line ln = intercalate "," $ map tr_cell ln
    tr_cell cl
      | any (`elem` "\"\n\r,") cl = quote cl
      | otherwise = cl
    quote x = "\"" ++ concatMap doubleDoubleQuote x ++ "\""
    doubleDoubleQuote '"' = "\"\""
    doubleDoubleQuote c = [c]
