# include <exception>
# include <sstream>
# include <iostream>

# include "boost/optional.hpp"
# include "boost/format.hpp"

# include "internal.h"

using namespace m100sim;
using boost::format;

//# define MAYBE_NOINLINE __attribute__((noinline)) /* for profiling */
# define MAYBE_NOINLINE

namespace
{

typedef m100sim_match_result mresult;

void show_team_state(const tstate &ts, logger &lg)
{
  for(list<cstate>::const_iterator it = ts.chs.begin(); it != ts.chs.end(); ++it)
    if(FILE *out = lg.start_direct())
    {
      print_cstate(out, *it);
      lg.end_direct();
    }
}

void show_current_state(const gstate &gs, logger &lg)
{
  bool multi = gs.teams[Left].chs.size() + gs.teams[Right].chs.size() > 2;
  if(multi) lg.put("");
  show_team_state(gs.teams[Left], lg);
  if(multi) lg.put("");
  show_team_state(gs.teams[Right], lg);
  lg.put("");
}

void reset_ap(gstate &gs, logger &lg)
{
  for(int i = 0; i < 2; i++)
    for(list<cstate>::iterator it = gs.teams[i].chs.begin(); it != gs.teams[i].chs.end(); ++it)
      if(it->bad.sleep > 0)
      {
        lg.putf("%sは眠っている", it->info.name.c_str());
        it->ap = 0;
        it->reserved_ap = 0;
        it->bad.sleep--;
      }
      else
        it->ap = it->spd;
  gs.move_next = gs.action_order.begin();
}

enum action_type
{
  at_sword,
  at_nop,
  at_stun
};

struct action
{
  effect_context ctx;
  action_type type;
  optional<sstate> sword;
  int prio;
};

bool usable(int turn, const cstate *cs, const sstate &ss)
{
  return !ss.used && turn >= ss.info->turn - ss.ench.early 
    && (!ss.info->mp || *ss.info->mp <= cs->mp)
    && (!(ss.info->reacts_to_danger || ss.ench.reacts_to_danger) || cs->hp <= cs->max_hp / 2)
    && (!ss.info->reacts_to_illness || !cs->bad.empty());
}

bool choose_action(gstate &gs, cstate *cs, action *out, logger &lg) MAYBE_NOINLINE;
bool choose_action(gstate &gs, cstate *cs, action *out, logger &lg)
{
  out->ctx.owner = cs;
  if(cs->bad.stun > 0)
  {
    cs->bad.stun--;
    out->type = at_stun;
    out->prio = prio_skp;
    return true;
  }
  for(sref it = cs->swords.begin(); it != cs->swords.end(); ++it)
    if(usable(gs.turn, cs, *it))
    {
      out->sword = *it;
      out->ctx.this_sword = &*it;
      out->ctx.ench = it->ench;
      sref next = it;
      if(++next != cs->swords.end())
        out->ctx.next_sword = next;
      out->type = at_sword;
      out->prio = it->info->prio;

      /* pay costs */
      if(it->info->mp)
        cs->mp -= *it->info->mp;
      if(it->info->limited || it->ench.limited)
      {
        cs->swords.erase(it);
        out->ctx.this_sword = NULL;
      }
      else
        it->used = true;
      return true;
    }
  out->type = at_nop;
  out->prio = prio_skp;
  return false;
}

void eliminate_dead(gstate &gs, logger &lg)
{
  if(!gs.someone_may_be_dead) return;
  list<cstate> to_remove;
  for(int i = 0; i < 2; i++)
    for(list<cstate>::iterator it = gs.teams[i].chs.begin(); it != gs.teams[i].chs.end();)
      if(it->dead)
      {
        list<cstate>::iterator next = it;
        ++next;
        unregister_from_action_order(gs, &*it);
        to_remove.splice(to_remove.end(), gs.teams[i].chs, it);
        it = next;
      }
      else
        ++it;
  if(to_remove.empty()) return;
  stringstream msg;
  bool first = true;
  for(list<cstate>::iterator it = to_remove.begin(); it != to_remove.end(); ++it)
  {
    if(!first) msg << "、";
    msg << it->info.name;
    tstate &ts = gs.teams[it->info.side];
    ts.hp_removed += it->hp;
    ts.push_removed += it->other.push;
    first = false;
  }
  msg << "は退場した";
  lg.put(msg.str());
}

bool gameover(const gstate &gs, logger &lg)
{
  const char *reason = NULL;
  if(gs.turn > 20)
    reason = "ターンオーバー";
  else if(gs.teams[0].chs.empty() || gs.teams[1].chs.empty())
    reason = "全滅";
  if(!reason) return false;
  lg.put("");
  lg.putf("%sのため戦闘終了", reason);
  return true;
}

mresult judge(const gstate &gs, logger &lg)
{
  ostringstream msg;
  mresult res;
  bool e_left = gs.teams[0].chs.empty();
  bool e_right = gs.teams[1].chs.empty();
  if(e_left && !e_right)
  {
    msg << "相手側のみ全滅";
    res = m100sim_right_wins;
  }
  else if(!e_left && e_right)
  {
    msg << "相手側のみ全滅";
    res = m100sim_left_wins;
  }
  else
  {
    int push[2] = {0};
    int_t hp[2];
    hp[0] = 0;
    hp[1] = 0;
    if(e_left)
    {
      msg << "両者全滅";
      for(int i = 0; i < 2; i++)
      {
        push[i] = gs.teams[i].push_removed;
        hp[i] = gs.teams[i].hp_removed;
      }
    }
    else
    {
      msg << "両パーティ生存";
      for(int i = 0; i < 2; i++)
        for(list<cstate>::const_iterator it = gs.teams[i].chs.begin(); it != gs.teams[i].chs.end(); ++it)
        {
          push[i] += it->other.push;
          hp[i] += it->hp;
        }
    }
    if(push[0] != push[1])
      msg << "、プッシュ" << push[0] << "対" << push[1];
    else if(push[0] != 0)
      msg << "、両者プッシュ" << push[0];
    if(push[0] > push[1])
      res = m100sim_left_wins;
    else if(push[0] < push[1])
      res = m100sim_right_wins;
    else
    {
      msg << "、HP" << hp[0] << "対" << hp[1];
      if(hp[0] > hp[1])
        res = m100sim_left_wins;
      else if(hp[0] < hp[1])
        res = m100sim_right_wins;
      else
        res = m100sim_draw;
    }
  }
  msg << "により、";
  switch(res)
  {
  case m100sim_left_wins:
    msg << gs.teams[0].name << "が勝利した";
    break;
  case m100sim_right_wins:
    msg << gs.teams[1].name << "が勝利した";
    break;
  case m100sim_draw:
    msg << "引き分けとなった";
    break;
  }
  lg.put(msg.str());
  return res;
}

boost::optional<mresult> checkpoint(gstate &gs, logger &lg)
{
  eliminate_dead(gs, lg);
  if(gameover(gs, lg))
    return judge(gs, lg);
  return boost::none;
}

void declare_actions(FILE *out, const vector<action> &acts)
{
  for(int i = 0; i < int(acts.size() - 1); i++)
  {
    fputs(acts[i].ctx.owner->info.name.c_str(), out);
    switch(acts[i].type)
    {
    case at_nop:
      fputs("は何もせず、", out);
      break;
    case at_sword:
      fputs("は", out);
      print_sstate(out, *acts[i].sword, false);
      fputs("を使い、", out);
      break;
    case at_stun:
      fputs("は気絶しており、", out);
      break;
    }
  }
  fputs(acts.back().ctx.owner->info.name.c_str(), out);
  switch(acts.back().type)
  {
  case at_nop:
    fputs("は何もしない", out);
    break;
  case at_sword:
    fputs("は", out);
    print_sstate(out, *acts.back().sword, false);
    fputs("を使った", out);
    break;
  case at_stun:
    fputs("は気絶している", out);
    break;
  }
}

void everyone_does_nothing(gstate &gs, const vector<action> &acts, logger &lg)
{
  int_t count = acts.front().ctx.owner->ap;
  for(int i = 1; i < int(acts.size()); i++)
    count = min(count, acts[i].ctx.owner->ap);
  for(int i = 0; i < int(acts.size()); i++)
    acts[i].ctx.owner->ap -= count;
  if(FILE *out = lg.start_direct())
  {
    declare_actions(out, acts);
    if(count > 0)
    {
      fputs("。", out);
      print_int(out, count + 1);
      fputs("回。", out);
    }
    lg.end_direct();
  }
}

struct te_info
{
  const effect_context *ctx;
  const rich_effect *re;
  cstate *target;
};

void process_prim(gstate &gs, const te_info *is, int ni, logger &lg) MAYBE_NOINLINE;
void process_prim(gstate &gs, const te_info *is, int ni, logger &lg)
{
  auto_ptr<prim_effect_base> pes[2];
  for(int i = 0; i < ni; i++)
    pes[i] = auto_ptr<prim_effect_base>(is[i].re->gen());
  while(true)
  {
    bool active = false;
    for(int i = 0; i < ni; i++)
      if(pes[i].get())
      {
        if(!pes[i]->perform(gs, *is[i].ctx, is[i].target, lg))
          pes[i].reset();
        active = true;
      }
    if(!active) break;
  }
}

void process_defended(gstate &gs, const te_info *is, int ni, logger &lg) MAYBE_NOINLINE;
void process_defended(gstate &gs, const te_info *is, int ni, logger &lg)
{
  /* shortcut */
  if((!is[0].re->defendable || is[0].target->defence.empty()) && (ni == 1 || !is[1].re->defendable || is[1].target->defence.empty()))
  {
    process_prim(gs, is, ni, lg);
    return;
  }
  cstate *targets[2];
  cstate *antitargets[2];
  te_info info[2];
  indent_controller ic(lg);
  for(int i = 0; i < ni; i++)
  {
    targets[i] = is[i].target;
    antitargets[i] = is[i].ctx->owner;
  }
  while(true)
  {
    int n = 0;
    bool effect_fired = false;
    ostringstream msg;
    for(int i = 0; i < ni; i++)
      if(!targets[i])
        continue;
      else if(!is[i].re->defendable || targets[i]->defence.empty())
      {
        info[n].ctx = is[i].ctx;
        info[n].re = is[i].re;
        info[n].target = targets[i];
        n++;
        targets[i] = NULL;
      }
      else
      {
        multiset<defence_status>::iterator fst = targets[i]->defence.begin();
        defence_status s = *fst;
        if(effect_fired) msg << "と";
        msg << targets[i]->info.name << "の" << s;
        targets[i]->defence.erase(fst);
        switch(s)
        {
        case reflection:
          swap(targets[i], antitargets[i]);
          break;
        case detection:
          targets[i] = NULL;
          break;
        }
        effect_fired = true;
      }
    if(n == 0 && !effect_fired) break;
    if(n > 0) process_prim(gs, info, n, lg);
    if(effect_fired)
    {
      msg  << "が発動した";
      lg.put(msg.str());
    }
    ic.add(1);
  }
}

void process_repeated(gstate &gs, const te_info *is, int ni, logger &lg) MAYBE_NOINLINE;
void process_repeated(gstate &gs, const te_info *is, int ni, logger &lg)
{
  te_info info[2];
  for(int j = 0; ; j++)
  {
    int n = 0;
    for(int i = 0; i < ni; i++)
      if(is[i].re->count > j)
        info[n++] = is[i];
    if(n == 0) break;
    if(!small_enough(gs.action_count += 2))
      throw danger_of_overflow();
    process_defended(gs, info, n, lg);
  }
}

typedef pair<const effect_context *, const rich_effect *> re_info;

typedef list<cstate>::iterator lc_iter;

cstate *choose_attack_target(gstate &gs, side_id side) MAYBE_NOINLINE;
cstate *choose_attack_target(gstate &gs, side_id side)
{
  list<cstate> &t = gs.teams[side].chs;
  for(list<cstate>::iterator it = t.begin(); it != t.end(); ++it)
    if(it->other.provocation)
      return &*it;
  return &t.front();
}

void setup_target(gstate &gs, cstate **ats, const re_info &ri, cstate **one, lc_iter *begin, lc_iter *end) MAYBE_NOINLINE;
void setup_target(gstate &gs, cstate **ats, const re_info &ri, cstate **one, lc_iter *begin, lc_iter *end)
{
  bool scatter = ri.second->scatter || ri.first->ench.scatter;
  cstate *self = ri.first->owner;
  side_id side = self->info.side;
  switch(ri.second->type)
  {
  case target_enemy:
  {
    side_id opp = opponent(side);
    if(scatter)
    {
      list<cstate> &tgt = gs.teams[opp].chs;
      *begin = tgt.begin();
      *end = tgt.end();
      *one = NULL;
    }
    else
    {
      /* target decision cannot be simply delayed because someone may have gained attraction
      if(!ats[opp])
        ats[opp] = choose_attack_target(gs, opp);
      */
      *one = ats[opp];
    }
    break;
  }
  case target_ally:
    if(scatter)
    {
      list<cstate> &tgt = gs.teams[side].chs;
      *begin = tgt.begin();
      *end = tgt.end();
      *one = NULL;
    }
    else
      *one = self;
    break;
  case target_none:
    *one = self;
    break;
  }
  if(*one)
    *begin = *end = gs.teams[0].chs.end();
}

void process_targeted(gstate &gs, cstate **ats, const re_info *is, int ni, logger &lg) MAYBE_NOINLINE;
void process_targeted(gstate &gs, cstate **ats, const re_info *is, int ni, logger &lg)
{
  cstate *one[2];
  lc_iter begins[2];
  lc_iter ends[2];
  te_info info[2];
  for(int i = 0; i < ni; i++)
    setup_target(gs, ats, is[i], one + i, begins + i, ends + i);
  while(true)
  {
    int n = 0;
    for(int i = 0; i < ni; i++)
      if(one[i])
      {
        info[n].ctx = is[i].first;
        info[n].re = is[i].second;
        info[n].target = one[i];
        n++;
        one[i] = NULL;
      }
      else if(begins[i] != ends[i])
      {
        info[n].ctx = is[i].first;
        info[n].re = is[i].second;
        info[n].target = &*begins[i];
        n++;
        ++begins[i];
      }
    if(n == 0) break;
    process_repeated(gs, info, n, lg);
  }
}

void process_independent(gstate &gs, cstate **ats, const action *const *as, int na, logger &lg) MAYBE_NOINLINE;
void process_independent(gstate &gs, cstate **ats, const action *const *as, int na, logger &lg)
{
  re_info info[2];
  for(int j = 0; ; j++)
  {
    int n = 0;
    for(int i = 0; i < na; i++)
      if(j < as[i]->sword->info->eff.size())
      {
        info[n].first = &as[i]->ctx;
        info[n].second = &as[i]->sword->info->eff[j];
        n++;
      }
    if(n == 0) break;
    process_targeted(gs, ats, info, n, lg);
  }
}

void process_single(gstate &gs, cstate **ats, bool multi, const action &x, logger &lg)
{
  if(x.type != at_sword) return;
  indent_controller ic(lg);
  if(multi)
  {
    lg.putf("処理: %sの%s", x.ctx.owner->info.name.c_str(), x.sword->info->name);
    ic.add(1);
  }
  const action *p = &x;
  process_independent(gs, ats, &p, 1, lg);
}

void process_supersimul(gstate &gs, cstate **ats, const action &x, const action &y, logger &lg)
{
  if(x.type != at_sword) return;
  lg.putf("%sの%sと%sの%sとの超同時処理",
     x.ctx.owner->info.name.c_str(), x.sword->info->name, y.ctx.owner->info.name.c_str(), y.sword->info->name);
  indent_controller ic(lg, 1);
  const action *as[2];
  as[0] = &x;
  as[1] = &y;
  process_independent(gs, ats, as, 2, lg);
}

void simul_action(gstate &gs, const vector<action> &acts, logger &lg) MAYBE_NOINLINE;
void simul_action(gstate &gs, const vector<action> &acts, logger &lg)
{
  if(FILE *out = lg.start_direct())
  {
    declare_actions(out, acts);
    lg.end_direct();
  }

  indent_controller ic(lg, 1);
  cstate *attack_targets[2] = {NULL, NULL};

  /* FIXME: should be lazy */
  attack_targets[Left] = choose_attack_target(gs, Left);
  attack_targets[Right] = choose_attack_target(gs, Right);

  /* shortcuts */
  if(acts.size() == 1)
  {
    process_single(gs, attack_targets, false, acts[0], lg);
    return;
  }
  else if(acts.size() == 2 && acts[0].ctx.owner->info.side != acts[1].ctx.owner->info.side)
  {
    if(acts[0].prio == acts[1].prio)
      process_supersimul(gs, attack_targets, acts[0], acts[1], lg);
    else if(acts[0].prio < acts[1].prio)
    {
      process_single(gs, attack_targets, true, acts[1], lg);
      process_single(gs, attack_targets, true, acts[0], lg);
    }
    else
    {
      process_single(gs, attack_targets, true, acts[0], lg);
      process_single(gs, attack_targets, true, acts[1], lg);
    }
    return;
  }

  vector<vector<const action *> > sorted;
  sorted.resize(n_priority);

  for(int i = 0; i < acts.size(); i++)
    sorted[acts[i].prio].push_back(&acts[i]);
  for(int i = n_priority - 1; i >= 0; i--)
  {
    int m;
    for(m = 0; m < sorted[i].size(); m++)
      if(sorted[i][m]->ctx.owner->info.side == Right)
        break;
    int l = 0, r = m;
    for(; l < m && r < sorted[i].size(); l++, r++)
      process_supersimul(gs, attack_targets, *sorted[i][l], *sorted[i][r], lg);
    for(; l < m; l++)
      process_single(gs, attack_targets, true, *sorted[i][l], lg);
    for(; r < sorted[i].size(); r++)
      process_single(gs, attack_targets, true, *sorted[i][r], lg);
  }
}

bool can_do_something(const gstate &gs, cstate *cs)
{
  if(cs->bad.stun > 0)
    return true;
  for(sref it = cs->swords.begin(); it != cs->swords.end(); ++it)
    if(usable(gs.turn, cs, *it))
      return true;
  return false;
}

# define RETURN_IF_SOME(expr) do{ boost::optional<mresult> r = expr; if(r) return r; } while(false)

boost::optional<mresult> do_action(gstate &gs, logger &lg, bool *more_action) MAYBE_NOINLINE;
boost::optional<mresult> do_action(gstate &gs, logger &lg, bool *more_action)
{
  vector<cstate *> active;
  bool empty = true;
  int mq;
  int_t mspd;

  active.reserve(gs.teams[Left].chs.size() + gs.teams[Right].chs.size());

  for(int i = 0; i < 2; i++)
    for(list<cstate>::iterator it = gs.teams[i].chs.begin(); it != gs.teams[i].chs.end(); ++it)
      if(it->ap > 0 || it->reserved_ap > 0 && can_do_something(gs, &*it))
      {
        int q = it->other.quick - it->other.slow;
        if(empty || q > mq || q == mq && it->spd > mspd)
        {
          active.clear();
          active.push_back(&*it);
          empty = false;
          mq = q;
          mspd = it->spd;
        }
        else if(q == mq && it->spd == mspd)
          active.push_back(&*it);
      }

  if(empty)
  {
    *more_action = false;
    return boost::none;
  }
 
  vector<action> acts;
  acts.resize(active.size());
  bool something_to_do = false;

  for(int i = 0; i < active.size(); i++)
  {
    if(active[i]->ap > 0)
      active[i]->ap--;
    else
      active[i]->reserved_ap--;
    if(choose_action(gs, active[i], &acts[i], lg))
      something_to_do = true;
  }

  if(something_to_do)
  {
    simul_action(gs, acts, lg);
    gs.sword_trash.clear();
    return checkpoint(gs, lg);
  }
  else
  {
    everyone_does_nothing(gs, acts, lg);
    return boost::none;
  }
}

optional<mresult> status_effect_phase(gstate &gs, logger &lg) MAYBE_NOINLINE;
optional<mresult> status_effect_phase(gstate &gs, logger &lg)
{
  for(int i = 0; i < 2; i++)
    for(list<cstate>::iterator ci = gs.teams[i].chs.begin(); ci != gs.teams[i].chs.end(); ++ci)
    {
      int bad = ci->bad.bad_poison;
      ci->bad.poison += bad;
      for(int k = 0; k < bad; k++) lg.putf("%sは猛毒によって毒を得た", ci->info.name.c_str());
    }
  for(int i = 0; i < 2; i++)
    for(list<cstate>::iterator ci = gs.teams[i].chs.begin(); ci != gs.teams[i].chs.end(); ++ci)
      if(ci->bad.poison)
        damage(gs, "毒", &*ci, ci->bad.poison, lg);
  return checkpoint(gs, lg);
}

void cleanup(gstate &gs, logger &lg) MAYBE_NOINLINE;
void cleanup(gstate &gs, logger &lg)
{
  for(int i = 0; i < 2; i++)
    for(list<cstate>::iterator ci = gs.teams[i].chs.begin(); ci != gs.teams[i].chs.end(); ++ci)
      for(sref s = ci->swords.begin(); s != ci->swords.end(); ++s)
        s->used = false;
  lg.put("");
}

boost::optional<mresult> turn(gstate &gs, logger &lg)
{
  RETURN_IF_SOME(checkpoint(gs, lg));
  if(!lg.null())
    show_current_state(gs, lg);
  RETURN_IF_SOME(status_effect_phase(gs, lg));
  reset_ap(gs, lg);
  bool more_action = true;
  do
  {
    RETURN_IF_SOME(do_action(gs, lg, &more_action));
  } while(more_action);
  cleanup(gs, lg);
  increment_turn_count(gs, lg);
  return boost::none;
}

} /* namespace */

mresult m100sim_match(m100sim_match_options opts, m100sim_character left, m100sim_character right) try
{
  logger lg(opts.log_output, opts.line_terminator);
  gstate gs(left, right);

  boost::optional<mresult> res;
  try
  {
    do
    {
      res = turn(gs, lg);
    } while(!res);
  } catch(danger_of_overflow &)
  {
    lg.put("*** オーバーフロー発生によりシミュレーション中断 ***");
    return m100sim_aborted;
  }
  return *res;
}
catch(exception &e)
{
  cerr << "m100sim_match: unexpected error: " << e.what() << "\n";
  return m100sim_aborted;
} catch(...)
{
  cerr << "m100sim_match: unexpected error: (no info)\n";
  return m100sim_aborted;
}
