{-# OPTIONS_GHC -fglasgow-exts #-}
{- |
   License      : BSD3
   Maintainer   : Takano Akio <aljee@hyper.cx>
   Stability    : experimental

   'scanner' transforms memory IO into stream IO.

   Some operations require element type to be an instance of
   'Typeable'. This is due to an internal performance hack.
-}
module System.IO.SSC.Scanner
  ( iScanner
  , oScanner
  , ioScanner
  , Scanner
  ) where

import Control.Monad
import Data.Typeable
import Data.Word
import Data.IORef
import Foreign.Storable
import Foreign.Ptr
import Foreign.Marshal.Array
import System.IO.Error
import System.IO.Unsafe

import System.IO.SSC.Internal.Ptr
import System.IO.SSC.Internal.Debug
import System.IO.SSC.Classes

newtype Scanner d pos a ba m = MS m

iScanner, oScanner, ioScanner :: m -> IO (Scanner d pos a ba m)
iScanner = scanner
oScanner = scanner
ioScanner = scanner

scanner :: m -> IO (Scanner d pos a ba m)
scanner mem = return $ MS mem

instance (Port m) => Port (Scanner d pos a ba m) where
  rawClose (MS mem) = rawClose mem
  rawIsClosed (MS mem) = rawIsClosed mem
  rawFlush (MS mem) = rawFlush mem

instance Destination d m => Destination d (Scanner d pos a ba m) where
  rawDest (MS mem) = rawDest mem

instance (Lockable m) => Lockable (Scanner d pos a ba m) where
  sLock (MS mem) = sLock mem
  sTryLock (MS mem) = sTryLock mem
  sUnsafeUnlock (MS mem) = sUnsafeUnlock mem

instance (Discardable m) => Discardable (Scanner d pos a ba m) where
  rawDiscard (MS mem) = rawDiscard mem

instance (Seekable pos m) => Seekable pos (Scanner d pos a ba m) where
  rawGetPos (MS mem) = rawGetPos mem
  rawSetPos (MS mem) = rawSetPos mem

instance (RelativeSeekable pos m) => RelativeSeekable pos (Scanner d pos a ba m) where
  rawRelativeSeek (MS mem) = rawRelativeSeek mem

instance (SizedInput pos m) => SizedInput pos (Scanner d pos a ba m) where
  rawInputEnd (MS mem) = rawInputEnd mem
  rawSeekFromEnd (MS mem) = rawSeekFromEnd mem

instance (SizedOutput pos m) => SizedOutput pos (Scanner d pos a ba m) where
  rawOutputEnd (MS mem) = rawOutputEnd mem

msGet :: (IMemory a m) => Scanner d pos a ba m -> IO a
msGet (MS mem) = do
  (p0, p1) <- rawReadMem mem
  if p0 == p1
    then do
      rawEndRead mem p0
      new <- rawFillMemB mem NonBlockingPreferred
      case new of
        Nothing -> ioError $ mkIOError eofErrorType "Scanner: rawGet" Nothing Nothing
        Just _ -> do
          (p0', p1') <- rawReadMem mem
          get p0'
    else get p0
  where
    get ptr = do
      x <- peek ptr
      rawEndRead mem (advancePtr ptr 1)
      return x

msGets :: (IMemory a m) => Scanner d pos a ba m -> Int -> IO [a]
msGets (MS mem) n
  | n <= 0 = return []
  | otherwise = do
    r <- loop n
    case r of
      [] -> ioError $ mkIOError eofErrorType "msGets" Nothing Nothing
      _ -> return r
  where
    loop 0 = return []
    loop n = do
      (p0, p1) <- rawReadMem mem
      let size = distancePtr p1 p0
      if size >= n
        then do
          let end = advancePtr p0 n
          r <- unpack p0 end
          rawEndRead mem end
          return r
        else do
          r <- unpack p0 p1
          rawEndRead mem p1
          new <- rawFillMemB mem NonBlockingPreferred
          case new of
            Nothing -> return r
            Just _ -> do
              rest <- loop (n - size)
              return (r ++ rest)

    unpack p0 p1 = seq p0 $ loop p1 []
      where
        loop p1 xs
          | p0 == p1 = return xs
          | otherwise = do
            let p1' = advancePtr p1 (-1)
            x <- peek p1'
            loop p1' (x:xs)

type ReadMem a = IO (Ptr a, Ptr a)
type EndRead a = Ptr a -> IO ()
type Fill = IO (Maybe Int)

data InputAux a t = forall m. (IMemory t m) => IA m t (IO [t] -> IO [a])

trivialAux :: (IMemory a m) => m -> a -> InputAux a a
trivialAux mem end = IA mem end id

--{-# RULES "msGetUntil/Char" forall ms end. msGetUntil ms end = doGetUntil (getUntilLoopChar (unMS ms) end) #-}
{-# NOINLINE msGetUntil #-}
msGetUntil :: (IMemory a m, Typeable a, Eq a) => Scanner d pos a ba m -> a -> IO [a]
msGetUntil (MS mem) end = doGetUntil $ case (gcast aux, gcast aux) of
    (Just (IA mem' end' from), _) -> from . getUntilLoopChar mem' end'
    (_, Just (IA mem' end' from)) -> from . getUntilLoopWord8 mem' end'
    _ -> getUntilLoopOther mem end
  where
    aux = trivialAux mem end

doGetUntil :: (IORef Bool -> IO [a]) -> IO [a]
doGetUntil body = do
  seen_newline <- newIORef True
  r <- body seen_newline
  case r of
    [] -> do
      s <- readIORef seen_newline
      if s
        then return []
        else ioError $ mkIOError eofErrorType "Scanner: sGetUnitil" Nothing Nothing
    _ -> return r

-- GHC's SPECIALIZE pragma does not perform enough specialization (see bug #1045)
-- so specialize manually here

peekChar :: Ptr Char -> IO Char
peekChar = peek

nextPtrChar :: Ptr Char -> Ptr Char
nextPtrChar = flip advancePtr 1

prevPtrChar :: Ptr Char -> Ptr Char
prevPtrChar = flip advancePtr (-1)

peekWord8 :: Ptr Word8 -> IO Word8
peekWord8 = peek

nextPtrWord8 :: Ptr Word8 -> Ptr Word8
nextPtrWord8 = flip advancePtr 1

getUntilLoopOther :: (IMemory a m, Eq a) => m -> a -> IORef Bool -> IO [a]
getUntilLoopOther mem end ref = self
  where
    self = do
      (p0, p1) <- rawReadMem mem
      let
        loop p
          | p == p1 = do
            rawEndRead mem p
            new <- rawFillMemB mem NonBlockingPreferred
            case new of
              Nothing -> do
                writeIORef ref False
                return []
              Just _ -> self
          | otherwise = do
            c <- peek p
            if c == end
              then do
                rawEndRead mem (advancePtr p 1)
                return []
              else do
                rest <- loop (advancePtr p 1)
                return (c:rest)
      p1 `seq` loop p0

findChar :: Ptr Char -> Ptr Char -> Char -> IO (Ptr Char)
findChar first last ch
  | ch `seq` False = undefined
  | first == last = return last
  | otherwise = do
    c <- peekChar first
    if c == ch
      then return first
      else findChar (nextPtrChar first) last ch

prependChars :: Ptr Char -> Ptr Char -> [Char] -> IO [Char]
prependChars first cur body
  | first > cur = return body
  | otherwise = do
    c <- peekChar cur
    prependChars first (prevPtrChar cur) (c:body)

getUntilLoopChar :: (IMemory Char m) => m -> Char -> IORef Bool -> IO [Char]
getUntilLoopChar mem end ref
  | mem `seq` end `seq` ref `seq` False = undefined
getUntilLoopChar mem end ref = do
  (p0, p1) <- rawReadMem mem
  endp <- findChar p0 p1 end
  if endp /= p1
    then do
      r <- prependChars p0 (prevPtrChar endp) []
      rawEndRead mem (nextPtrChar endp)
      return r
    else do
      ph <- newIORef (error "getUntilLoop: placeholder evaluated")
      cont <- unsafeInterleaveIO $ readIORef ph
      r <- prependChars p0 (prevPtrChar endp) cont
      rawEndRead mem endp
      new <- rawFillMemB mem NonBlockingPreferred
      case new of
        Nothing -> do
          writeIORef ph []
          writeIORef ref False
          return r
        Just _ -> do
          rest <- getUntilLoopChar mem end ref
          writeIORef ph rest
          return r

{-
getUntilLoopChar :: (IMemory Char m) => m -> Char -> IORef Bool -> IO [Char]
getUntilLoopChar mem end ref
  | mem `seq` end `seq` ref `seq` False = undefined
getUntilLoopChar mem end ref = do
  (p0, p1) <- rawReadMem mem
  let
    loop p
      | p == p1 = do
        rawEndRead mem p
        new <- rawFillMemB mem NonBlockingPreferred
        case new of
          Nothing -> do
            writeIORef ref False
            return []
          Just _ -> getUntilLoopChar mem end ref
      | otherwise = do
        c <- peekChar p
        if c == end
          then do
            rawEndRead mem (nextPtrChar p)
            return []
          else do
            rest <- loop (nextPtrChar p)
            return (c:rest)
  p1 `seq` loop p0
  -}

getUntilLoopWord8 :: (IMemory Word8 m) => m -> Word8 -> IORef Bool -> IO [Word8]
getUntilLoopWord8 mem end ref
  | mem `seq` end `seq` ref `seq` False = undefined
getUntilLoopWord8 mem end ref = self
  where
    self = do
      (p0, p1) <- rawReadMem mem
      let
        loop p
          | p == p1 = do
            rawEndRead mem p
            new <- rawFillMemB mem NonBlockingPreferred
            case new of
              Nothing -> do
                writeIORef ref False
                return []
              Just _ -> self
          | otherwise = do
            c <- peekWord8 p
            if c == end
              then do
                rawEndRead mem (nextPtrWord8 p)
                return []
              else do
                rest <- loop (nextPtrWord8 p)
                return (c:rest)
      p1 `seq` loop p0

data GetContentsAux a t = GA (ReadMem t) (EndRead t) (IO [t] -> IO [a])

msGetContents :: (Port m, IMemory a m, Typeable a) => Scanner d pos a ba m -> IO [a]
msGetContents ms@(MS mem) = case (gcast aux, gcast aux) of
  (Just (GA readMem' endRead' from), _) -> from $ getContentsChar readMem' endRead' fill close
  (_, Just (GA readMem' endRead' from)) -> from $ getContentsWord8 readMem' endRead' fill close
  _ -> getContentsOther readMem endRead fill close
  where
    aux = GA readMem endRead id
    readMem = rawReadMem mem
    endRead = rawEndRead mem
    fill = rawFillMemB mem NonBlockingPreferred
    close = rawClose mem

getContentsOther :: (Storable a) => ReadMem a -> EndRead a -> Fill -> IO () -> IO [a]
getContentsOther readMem endRead fill close = unsafeInterleaveIO loop
  where
    loop = do
      (p0, p1) <- readMem
      rest <- unsafeInterleaveIO $ do
        r <- try fill
        case r of
          (Right (Just _)) -> loop
          _ -> do
            close
            return []
      r <- loop1 p0 p1 rest
      endRead p1
      return r
    loop1 p0 p1 cs
      | p0 == p1 = return cs
      | otherwise = do
        let prev = advancePtr p1 (-1)
        c <- peek prev
        loop1 p0 prev (c:cs)

getContentsChar :: ReadMem Char -> EndRead Char -> Fill -> IO () -> IO [Char]
getContentsChar readMem endRead fill close = unsafeInterleaveIO loop
  where
    loop = do
      (p0, p1) <- readMem
      rest <- unsafeInterleaveIO $ do
        r <- try fill
        case r of
          (Right (Just _)) -> loop
          _ -> do
            close
            return []
      r <- loop1 p0 p1 rest
      endRead p1
      return r
    loop1 p0 p1 cs
      | p0 == p1 = return cs
      | otherwise = do
        let prev = advancePtr p1 (-1)
        c <- peek prev
        loop1 p0 prev (c:cs)

getContentsWord8 :: ReadMem Word8 -> EndRead Word8 -> Fill -> IO () -> IO [Word8]
getContentsWord8 readMem endRead fill close = unsafeInterleaveIO loop
  where
    loop = do
      (p0, p1) <- readMem
      rest <- unsafeInterleaveIO $ do
        r <- try fill
        case r of
          (Right (Just _)) -> loop
          _ -> do
            close
            return []
      r <- loop1 p0 p1 rest
      endRead p1
      return r
    loop1 p0 p1 cs
      | p0 == p1 = return cs
      | otherwise = do
        let prev = advancePtr p1 (-1)
        c <- peek prev
        loop1 p0 prev (c:cs)

instance (IMemory a m, Typeable a, Port m) => IStream a (Scanner d pos a ba m) where
  rawGet  = msGet
  rawGetContents = msGetContents
  rawGets = msGets

instance (IMemory a m, Typeable a, Port m, Eq a) => RetrievableStream a (Scanner d pos a ba m) where
  rawGetUntil = msGetUntil

ioeStuck = ioError $ mkIOError fullErrorType "Scanner: Put" Nothing Nothing

{-# INLINE msPut #-}
msPut :: (OMemory a m) => Scanner d pos a ba m -> a -> IO ()
msPut (MS mem) x = do
  (p0, p1) <- rawWriteMem mem
  if p0 == p1
    then do
      rawEndWrite mem p0
      (_, stuck) <- rawFlushMemB mem NonBlockingPreferred
      if stuck
        then ioeStuck
        else do
          (p0', p1') <- rawWriteMem mem
          put p0' p1'
    else put p0 p1
  where
    put p0 p1 = do
      poke p0 x
      rawEndWrite mem (advancePtr p0 1)

type WriteMem a = IO (Ptr a, Ptr a)
type EndWrite a = Ptr a -> IO ()
type Flush = IO (Int, Bool)
data OutputAux a = OA (WriteMem a) (EndWrite a)
newtype Putter a = Putter (WriteMem a -> EndWrite a -> Flush -> [a] -> IO ())

unMS :: Scanner d pos a ba m -> m
unMS (MS mem) = mem

--{-# RULES "msPuts/Char" forall ms. msPuts ms = putsChar (rawWriteMem (unMS ms)) (rawEndWrite (unMS ms)) (rawFlushMemB (unMS ms) NonBlockingPreferred) #-}
{-# SPECIALIZE msPuts :: (OMemory Char m) => Scanner d pos Char ba m -> [Char] -> IO () #-}
{-# NOINLINE msPuts #-}

msPuts :: (OMemory a m, Typeable a) => Scanner d pos a ba m -> [a] -> IO ()
msPuts (MS mem) xs = putter writeMem endWrite flush xs
  where
    putter = case gcast (Putter putsChar) of
      Just (Putter x) -> x
      Nothing -> case gcast (Putter putsWord8) of
        Just (Putter x) -> x
        Nothing -> putsOther
    writeMem = rawWriteMem mem
    endWrite = rawEndWrite mem
    flush = rawFlushMemB mem NonBlockingPreferred

putsOther :: (Storable a) => WriteMem a -> EndWrite a -> Flush -> [a] -> IO ()
putsOther writeMem endWrite flush = self
  where
    self xs = do
      (p0, p1) <- writeMem
      let
        loop p xs
          | p `seq` False = undefined
        loop p [] = endWrite p
        loop p xxs@(x:xs)
          | p == p1 = do
            endWrite p
            (_, stuck) <- flush
            if stuck
              then ioeStuck
              else self xxs
          | otherwise = do
            poke p x
            loop (advancePtr p 1) xs
      p1 `seq` loop p0 xs

putsChar :: WriteMem Char -> EndWrite Char -> Flush -> [Char] -> IO ()
putsChar writeMem endWrite flush = self
  where
    self xs = do
      (p0, p1) <- writeMem
      let
        loop p xs
          | p `seq` False = undefined
        loop p [] = endWrite p
        loop p xxs@(x:xs)
          | p == p1 = do
            endWrite p
            (_, stuck) <- flush
            if stuck
              then ioeStuck
              else self xxs
          | otherwise = do
            poke p x
            loop (advancePtr p 1) xs
      p1 `seq` loop p0 xs

putsWord8 :: WriteMem Word8 -> EndWrite Word8 -> Flush -> [Word8] -> IO ()
putsWord8 writeMem endWrite flush = self
  where
    self xs = do
      (p0, p1) <- writeMem
      let
        loop p xs
          | p `seq` False = undefined
        loop p [] = endWrite p
        loop p xxs@(x:xs)
          | p == p1 = do
            endWrite p
            (_, stuck) <- flush
            if stuck
              then ioeStuck
              else self xxs
          | otherwise = do
            poke p x
            loop (advancePtr p 1) xs
      p1 `seq` loop p0 xs

instance (OMemory a m, Typeable a) => OStream a (Scanner d pos a ba m) where
  rawPut = msPut
  rawPuts = msPuts

instance (IBlock ba m) => IBlock ba (Scanner d pos a ba m) where
  rawGetBufB (MS mem) = rawGetBufB mem

instance (OBlock ba m) => OBlock ba (Scanner d pos a ba m) where
  rawPutBufB (MS mem) = rawPutBufB mem

instance (IMemory a m) => IMemory a (Scanner d pos a ba m) where
  rawReadMem (MS mem) = rawReadMem mem
  rawEndRead (MS mem) = rawEndRead mem
  rawFillMemB (MS mem) = rawFillMemB mem

instance (OMemory a m) => OMemory a (Scanner d pos a ba m) where
  rawWriteMem (MS mem) = rawWriteMem mem
  rawEndWrite (MS mem) = rawEndWrite mem
  rawFlushMemB (MS mem) = rawFlushMemB mem
