{-# OPTIONS_GHC -fffi #-}
{- |
   License      : BSD3
   Maintainer   : Takano Akio <aljee@hyper.cx>
   Stability    : experimental

   Conversion between Win32 multibyte string and wide character
   string. Operations provided by this module demand codepage value
   to use. Only SBCS, DBCS and UTF-8 are supported by this module. In
   other words, encoding schemes in which one character can be
   represented by more than two bytes, such as UTF-7, are not
   supported except for UTF-8.
-}
module System.IO.SSC.Win32Unicode
  ( UnrepresentableCharHandler
  , mb2wConverter
  , w2mbConverter
  , iMb2wBlock
  , oMb2wBlock
  , ioMb2wBlock
  , oMb2wBlockWithHandler
  , ioMb2wBlockWithHandler
  , iW2mbBlock
  , oW2mbBlock
  , ioW2mbBlock
  , iW2mbBlockWithHandler
  , ioW2mbBlockWithHandler
  ) where

import Control.Monad
import Data.Word
import Foreign.ForeignPtr
import Foreign.C.Types
import Foreign.C.Error
import Foreign.Ptr
import Foreign.Storable
import System.IO.SSC.Classes
import System.IO.SSC.Convert
import System.IO.SSC.Internal.Debug
import System.IO.SSC.UTF8Block

foreign import ccall unsafe "win32_unicode.h win32_unicode_new_conv_d" c_new_conv_d :: CUInt -> IO (Ptr ())
foreign import ccall unsafe "win32_unicode.h &win32_unicode_delete_conv_d" c_delete_conv_d :: FunPtr (Ptr () -> IO ())
foreign import ccall unsafe "win32_unicode.h win32_unicode_char_to_wchar"
  c_char_to_wchar :: Ptr () -> Ptr CChar -> CSize -> Ptr CWchar -> CSize -> IO CInt
foreign import ccall unsafe "win32_unicode.h win32_unicode_wchar_to_char"
  c_wchar_to_char :: Ptr () -> Ptr CWchar -> CSize -> Ptr CChar -> CSize -> IO CInt
foreign import ccall unsafe "win32_unicode.h win32_unicode_last_read" c_last_read :: Ptr () -> IO CInt
foreign import ccall unsafe "win32_unicode.h win32_unicode_last_written" c_last_written :: Ptr () -> IO CInt

type ConvD = ForeignPtr ()
type UnrepresentableCharHandler = [Word16] -> Maybe [Word16]

newConvD :: Integer -> IO ConvD
newConvD cp = do
  p <- throwErrnoIfNull "System.IO.SSC.Win32Unicode.newConvD" $ c_new_conv_d (fromIntegral cp)
  newForeignPtr c_delete_conv_d p

cCharToWord16 :: ConvD -> Converter Word8 Word16
cCharToWord16 d iptr isize optr osize = withForeignPtr d $ \p -> do
  throwErrnoIfMinus1_ "System.IO.SSC.Win32Unicode.cCharToWord16" $
    c_char_to_wchar p (castPtr iptr) (fromIntegral isize) (castPtr optr) (fromIntegral osize)
  read <- c_last_read p
  written <- c_last_written p
  return (fromIntegral read, fromIntegral written)

cWcharToCChar :: ConvD -> UnrepresentableCharHandler -> Converter Word16 Word8
cWcharToCChar d f = converterWithFallback conv second
  where
    conv iptr isize optr osize = withForeignPtr d $ \p -> do
      r <- c_wchar_to_char p (castPtr iptr) (fromIntegral isize) (castPtr optr) (fromIntegral osize)
      succeed <- if r == -1
        then do
          errno <- getErrno
          when (errno /= eILSEQ) $ ioError $ userError "System.IO.SSC.Win32Unicode: error on wchar_to_char"
          return False
        else return True
      read <- c_last_read p
      written <- c_last_written p
      msg $ "last_read=" ++ show read ++ "; last_written=" ++ show written ++ "; succeed=" ++ show succeed
      return (fromIntegral read, fromIntegral written, succeed)

    second iptr isize
      | isize == 0 = return (0, [])
      | otherwise = do
      w0 <- peek iptr
      if isHighSurrogate w0
        then if isize == 1
          then return (0, [])
          else do
            w1 <- peekElemOff iptr 1
            cont [w0, w1]
        else cont [w0]
      where
        cont s = case f s of
          Nothing -> ioError $ userError "System.IO.SSC.Win32Unicode: Unrepresentable character"
          Just alt -> return (length s, alt)

    isHighSurrogate w = 0xd800 <= w && w < 0xdc00

cpUTF8 :: Integer
cpUTF8 = 65001

mb2wConverter :: Integer -> IO (Converter Word8 Word16)
mb2wConverter cp
  | cp == cpUTF8 = return utf8ToUTF16Converter
  | otherwise = do
    d <- newConvD cp
    return $ cCharToWord16 d

w2mbConverter :: Integer -> UnrepresentableCharHandler -> IO (Converter Word16 Word8)
w2mbConverter cp handler 
  | cp == cpUTF8 = return utf16ToUTF8Converter
  | otherwise = do
    d <- newConvD cp
    return $ cWcharToCChar d handler

ignoreHandler :: UnrepresentableCharHandler
ignoreHandler = const (Just [])

iMb2wBlock :: Integer -> m -> IO (IConversionBlock d pos m a Word8 Word16)
iMb2wBlock cp mem = mb2wConverter cp >>= flip iConversionBlock mem

oMb2wBlock :: Integer -> m -> IO (OConversionBlock d pos m a Word8 Word16)
oMb2wBlock cp mem = oMb2wBlockWithHandler cp ignoreHandler mem

ioMb2wBlock :: Integer -> m -> IO (IOConversionBlock d pos m a Word8 Word16)
ioMb2wBlock cp mem = ioMb2wBlockWithHandler cp ignoreHandler mem

oMb2wBlockWithHandler 
  :: Integer
  -> UnrepresentableCharHandler
  -> m
  -> IO (OConversionBlock d pos m a Word8 Word16)
oMb2wBlockWithHandler cp handler mem = w2mbConverter cp handler >>= flip oConversionBlock mem

ioMb2wBlockWithHandler
  :: Integer
  -> UnrepresentableCharHandler
  -> m
  -> IO (IOConversionBlock d pos m a Word8 Word16)
ioMb2wBlockWithHandler cp handler mem = do
  up <- mb2wConverter cp
  dn <- w2mbConverter cp handler
  ioConversionBlock up dn mem

iW2mbBlock :: Integer -> m -> IO (IConversionBlock d pos m a Word16 Word8)
iW2mbBlock cp mem = iW2mbBlockWithHandler cp ignoreHandler mem

oW2mbBlock :: Integer -> m -> IO (OConversionBlock d pos m a Word16 Word8)
oW2mbBlock cp mem = mb2wConverter cp >>= flip oConversionBlock mem

ioW2mbBlock :: Integer -> m -> IO (IOConversionBlock d pos m a Word16 Word8)
ioW2mbBlock cp mem = ioW2mbBlockWithHandler cp ignoreHandler mem

iW2mbBlockWithHandler
  :: Integer
  -> UnrepresentableCharHandler
  -> m
  -> IO (IConversionBlock d pos m a Word16 Word8)
iW2mbBlockWithHandler cp handler mem = w2mbConverter cp handler >>= flip iConversionBlock mem

ioW2mbBlockWithHandler
  :: Integer
  -> UnrepresentableCharHandler
  -> m
  -> IO (IOConversionBlock d pos m a Word16 Word8)
ioW2mbBlockWithHandler cp handler mem = do  
  up <- w2mbConverter cp handler
  dn <- mb2wConverter cp
  ioConversionBlock up dn mem
