
# include "win32_unicode.h"

# include <stdlib.h>
# include <stdio.h>
# include <errno.h>
# include <windows.h>

struct conv_d
{
  unsigned cp;
  int last_read;
  int last_written;
  int dbcs;
};

void *win32_unicode_new_conv_d(unsigned cp)
{
  CPINFO cpi;
  int is_dbcs;
  if(!GetCPInfo(cp, &cpi))
  {
    errno = EINVAL;
    return NULL;
  }
  switch(cpi.MaxCharSize)
  {
  case 1:
    is_dbcs = 0;
    break;
  case 2:
    is_dbcs = 1;
    break;
  default:
    errno = EINVAL;
    return NULL;
  }
  struct conv_d *d = malloc(sizeof(struct conv_d));
  if(!d) return NULL;
  d->cp = cp;
  d->dbcs = is_dbcs;
  return d;
}

void win32_unicode_delete_conv_d(void *p)
{
  free(p);
}

static int f_min(int x, int y)
{
  return x < y ? x : y;
}

int win32_unicode_char_to_wchar(void *p, const char *ip, size_t is, wchar_t *op, size_t os)
{
  struct conv_d *d = p;
  int read = 0, written = 0;
  int size = f_min(is, os);
  while(read < is && written < os && size)
  {
    int last;
    int r;
    if(d->dbcs)
    {
      if(size == 1 && is > read + 1 && IsDBCSLeadByteEx(d->cp, ip[read]))
        size = 2;
      else
      {
        for(last = read + size - 1; last >= read; last--)
          if(!IsDBCSLeadByteEx(d->cp, ip[last]))
            break;
        if(((read + size) - (last + 1)) % 2)
          if(!--size)
            break;
      }
    }
    r = MultiByteToWideChar(d->cp, MB_ERR_INVALID_CHARS, ip + read, size, op + written, os - written);
    if(r)
    {
      read += size;
      written += r;
      size = f_min(is - read, os - written);
    }
    else switch(GetLastError())
    {
    case ERROR_INSUFFICIENT_BUFFER:
      size = 1;
      break;
    case ERROR_NO_UNICODE_TRANSLATION:
      errno = EILSEQ;
      return -1;
    default:
      errno = EINVAL;
      return -1;
    }
  }
  d->last_read = read;
  d->last_written = written;
  return 0;
}

static int convert_until_unrepresentable(struct conv_d *, const wchar_t *, size_t, char *, size_t, int, int);

int win32_unicode_wchar_to_char(void *p, const wchar_t *ip, size_t is, char *op, size_t os)
{
  struct conv_d *d = p;
  int read = 0, written = 0;
  while(read < is && written < os)
  {
    BOOL err = 0;
    int r;
    int size = f_min(is - read, 
        d->dbcs && os - written != 1
          ? (os - written) / 2
          : os - written);
    r = WideCharToMultiByte(d->cp, 0, ip + read, size, op + written, os - written, NULL, &err);
    if(r)
    {
      if(err)
        return convert_until_unrepresentable(d, ip, is, op, os, read, written);
      read += size;
      written += r;
    }
    else if(GetLastError() == ERROR_INSUFFICIENT_BUFFER)
      break;
    else
    {
      errno = EINVAL;
      return -1;
    }
  }
  d->last_read = read;
  d->last_written = written;
  return 0;
}

static int convert_until_unrepresentable(
    struct conv_d *d, 
    const wchar_t *ip,
    size_t is,
    char *op,
    size_t os,
    int read,
    int written)
{
  while(read < is && written < os)
  {
    BOOL err = 0;
    int r = WideCharToMultiByte(d->cp, 0, ip + read, 1, op + written, os - written, NULL, &err);
    if(r)
    {
      if(err)
      {
        d->last_read = read;
        d->last_written = written;
        errno = EILSEQ;
        return -1;
      }
    }
    else if(GetLastError() == ERROR_INSUFFICIENT_BUFFER)
      break;
    else
    {
      errno = EINVAL;
      return -1;
    }
    read++;
    written += r;
  }
  d->last_read = read;
  d->last_written = written;
  return 0;
}

int win32_unicode_last_read(void *p)
{
  struct conv_d *d = p;
  return d->last_read;
}

int win32_unicode_last_written(void *p)
{
  struct conv_d *d = p;
  return d->last_written;
}

